import glob
import re
from collections import OrderedDict

import joblib
import numpy as np
from astropy import table
from astropy.io import fits
from astropy.time import Time

from .config import SONGCN_PARAMS


class SongCalendar(table.Table):
    """
    Initiated from all song observations.
    """

    def __init__(self, *args, **kwargs):
        super(SongCalendar, self).__init__(*args, **kwargs)

    @staticmethod
    def from_dir(root="/Users/cham/projects/song/star_spec", n_jobs=4):
        # glob files in root
        fps = glob.glob(root + "/*/night/raw/*.fits", recursive=True)
        fps.sort()
        print(f"{len(fps)} files found!")
        meta = OrderedDict(
            root=root,
            fps=fps,
        )

        # gather basic info
        basic_info = SongFile.gather_basic_info(fps, n_jobs=4)

        # stats for all IMAGETYP in each day
        date_list = np.unique(basic_info["jd1date"])
        info_list = []
        for date in date_list:
            od = OrderedDict(
                DATE=date,
                N_FILE=sum(basic_info["jd1date"] == date),
            )
            for k in SONGCN_PARAMS["IMAGETYP"]:
                od[k] = sum((basic_info["jd1date"] == date) & (basic_info["IMAGETYP"] == k))
            info_list.append(od)
        sc = SongCalendar(info_list, meta=meta)
        return sc

    def get_one_night(self, date="20191029"):
        night_list = list(self["DATE"])
        if date in night_list:
            basic_info = SongFile.gather_basic_info(self.meta["fps"], n_jobs=4)
            return SongNight(basic_info[basic_info["jd1date"] == date])
        else:
            print("Available DATEs are: ", list(self["DATE"]))
            raise ValueError(f"The given date {date} is not available!")


class SongNight(table.Table):
    def __init__(self, *args, **kwargs):
        super(SongNight, self).__init__(*args, **kwargs)

    # def from_dir(self, dir_path):
    #     """ from a night dir """
    #     pass

    def select(self,
               cond_dict={"IMAGETYP": "THAR", "SLIT": 6},
               method="random",
               n_select=120
               ):
        """ select some images from list

        Parameters
        ----------
        cond_dict: dict
            the dict of colname:value pairs
        method: str, {"all", "random", "top", "bottom"}
            the selection method
        n_select:
            the number of images that will be selected
            if n_images is larger than the number of images matched conditions,
            then n_images is forced to be n_matched

        Returns
        -------
        the Song instance

        Examples
        --------
        >>> s.list_image({"IMAGETYP":"STAR"}, returns=["OBJECT"])
        >>> s.select({"IMAGETYP":"THAR", "SLIT":6}, method="all", n_select=200,
        >>>          returns="sub", verbose=False)

        """

        # determine the matched images
        ind_match = np.ones((len(self),), dtype=bool)
        if cond_dict is None or len(cond_dict) < 1:
            print("@SONG: no condition is specified!")
        for k, v in cond_dict.items():
            ind_match = np.logical_and(ind_match, self[k] == v)

        # if no image found
        n_matched = np.sum(ind_match)
        if n_matched < 1:
            # print("@SONG: no images matched!")
            return []

        sub_match = np.where(ind_match)[0]
        # determine the number of images to select
        n_return = np.min([n_matched, n_select])

        # select according to method
        assert method in {"all", "random", "top", "bottom"}
        sub_rand = np.arange(0, n_matched, dtype=int)
        if method == "all":
            n_return = n_matched
        elif method == "random":
            np.random.shuffle(sub_rand)
            sub_rand = sub_rand[:n_return]
        elif method == "top":
            sub_rand = sub_rand[:n_return]
        elif method == "bottom":
            sub_rand = sub_rand[-n_return:]
        sub_return = sub_match[sub_rand]

        # constructing result to be returned
        return self["path"][sub_return]

    def master_bias(self):
        fps = self.select({"IMAGETYP":"THAR", "SLIT":6})
        return fps

    def master_flat(self):
        fps = self["path"][self["IMAGETYP"] == "FLAT"]
        return fps

    @staticmethod
    def catalog_spectra(file_paths):
        pass

    def get_image(self):
        pass


class SongFile:
    def __init__(self, path='/Users/cham/projects/song/star_spec/20191031/night/raw/s2_2019-10-31T17-22-11.fits'):
        super().__init__()
        m = re.search(
            r"(\d+)/night/raw/"
            r"s(?P<node>\d+)_(?P<yyyy>\d+)-(?P<mm>\d+)-(?P<dd>\d+)T(?P<HH>\d+)-(?P<MM>\d+)-(?P<SS>\d+).fits",
            path)
        header = fits.getheader(path)
        if m is not None:
            self.isregular = True
            self.path = path
            # self.node = m.group("node")
            # self.time = Time(
            #     "{}-{}-{}T{}:{}:{}".format(
            #         m.group("yyyy"), m.group("mm"), m.group("dd"), m.group("HH"), m.group("MM"), m.group("SS")
            #     ),
            #     format="isot"
            # )
        else:
            self.isregular = False
            self.path = path
        self.telescope = header["TELESCOP"]
        self.time = Time(header["JD-MID"], format="jd")
        self.jd1 = int(self.time.jd1)
        self.jd1date = Time(self.jd1, format="jd").to_datetime().strftime("%Y%m%d")

    def __getitem__(self, item):
        """ Keywords can be accessed in a simple way """
        header = fits.getheader(self.path)
        return header[item]

    @property
    def header(self):
        return fits.getheader(self.path)

    @property
    def data(self):
        return fits.getdata(self.path)

    def info(self):
        return fits.info(self.path)

    def basic_dict(self):
        """ Return basic info dict. """
        header = self.header
        return OrderedDict(
            path=self.path,
            isregular=self.isregular,
            telescope=self.telescope,
            jd1=self.jd1,
            jd1date=self.jd1date,
            jdmid=header["JD-MID"],
            SLIT=header["SLIT"],
            I2POS=header["I2POS"],
            IMAGETYP=header["IMAGETYP"],
            EXPTIME=header["EXPTIME"],
            OBJECT=header["OBJECT"],
            OBJNAME=header["OBJ-NAME"],
            OBJRA=header["OBJ-RA"],
            OBJDEC=header["OBJ-DEC"],
        )

    def full_dict(self):
        """ Return full info dict. """
        d = OrderedDict(
            path=self.path,
            isregular=self.isregular,
            telescope=self.telescope,
            jd1=self.jd1,
            jd1date=self.jd1date,
        )
        header = self.header
        for k in header.keys():
            if not k.startswith("---"):
                d[k] = header[k]
        return d

    @staticmethod
    def gather_basic_info(fps, n_jobs=4):
        def get_one_entry(path):
            return SongFile(path).basic_dict()

        return table.Table(
            joblib.Parallel(n_jobs=n_jobs)(
                joblib.delayed(get_one_entry)(_) for _ in fps
            )
        )

    @staticmethod
    def gather_full_table(fps, n_jobs=4):
        def get_one_entry(path):
            return SongFile(path).full_dict()

        return table.Table(
            joblib.Parallel(n_jobs=n_jobs)(
                joblib.delayed(get_one_entry)(_) for _ in fps
            )
        )


def test_songfile():
    sf = SongFile()
    print(sf["IMAGETYP"])
    print(sf.header)
    print(sf.data)
    sf.info()
    print(sf.basic_dict())
    print(sf.full_dict())


def test_songcalendar():
    sc = SongCalendar.from_dir()
    sc.pprint_all(align="<")
    sc.show_in_browser()


if __name__ == "__main__":
    test_songfile()



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Sun May 28 13:00:00 2017

Modifications
-------------
-

Aims
----
- routines to substract scattered light

"""

import logging

import numpy as np
from skimage import filters, morphology
from tqdm import tqdm


def straight_line_2pts(x, y):
    """ return coefs for a straight line: y = a*x+b, x and y are 2-element array """
    a = (y[1] - y[0]) / (x[1] - x[0])
    b = y[0] - a * x[0]
    return a, b


def background(img, al, q=(40, 5), median_sigma=15, gaussian_sigma=15):
    """
    Estimate background / scattered light using inter-aperture pixels.

    Parameters
    ----------
    img : ndarray
        the image whose background is to be determined
    al : ApertureList
        aperture center array, (n_aperture, n_pixel)
    q : tuple of float
        the starting and ending percentile
    median_sigma : int
        median smoothing parameter
    gaussian_sigma : int or tuple
        gaussian smoothing parameter

    Returns
    -------
    numpy.ndarray
        The estimated background / scattered light.
    """
    n_row, n_col = img.shape
    mask = al.get_mask(ap_width=10)  # interpolate for True values
    bg = np.where(mask, 0., img)
    qs = np.linspace(*q, num=len(al))

    # img_coord = np.arange(n_lambda)
    logging.info("Start processing background ...")
    for i_ap in tqdm(range(len(al)), unit="Aperture"):
        if len(al) == 1:
            ap_center_left = (al[i_ap].ap_center_floor - 2 * al[i_ap].ap_width).astype(int)
            ap_center_center = al[i_ap].ap_center_floor.astype(int)
            ap_center_right = (al[i_ap].ap_center_floor + 2 * al[i_ap].ap_width).astype(int)
        else:
            if i_ap == 0:
                ap_center_left = (2 * al[i_ap].ap_center_floor - al[i_ap + 1].ap_center_floor).astype(int)
                ap_center_center = al[i_ap].ap_center_floor.astype(int)
                ap_center_right = al[i_ap + 1].ap_center_floor.astype(int)
            elif i_ap == len(al) - 1:
                ap_center_left = al[i_ap - 1].ap_center_floor.astype(int)
                ap_center_center = al[i_ap].ap_center_floor.astype(int)
                ap_center_right = (2 * al[i_ap].ap_center_floor - al[i_ap - 1].ap_center_floor).astype(int)
            else:
                ap_center_left = al[i_ap - 1].ap_center_floor.astype(int)
                ap_center_center = al[i_ap].ap_center_floor.astype(int)
                ap_center_right = al[i_ap + 1].ap_center_floor.astype(int)

        pos_left = .5 * (ap_center_left + ap_center_center)
        pos_right = .5 * (ap_center_center + ap_center_right)

        for i_row in range(n_row):
            line_slope, line_intersect = straight_line_2pts(
                [
                    pos_left[i_row],
                    pos_right[i_row]
                ],
                [
                    np.percentile(img[i_row, max(0, ap_center_left[i_row]):max(1, ap_center_center[i_row])], qs[i_ap]),
                    np.percentile(img[i_row, max(0, ap_center_center[i_row]):max(1, ap_center_right[i_row])], qs[i_ap]),
                ]
            )
            interp_start = 0 if i_ap == 0 else max(0, ap_center_left[i_row])
            interp_stop = n_col if i_ap == len(al) - 1 else min(n_col, ap_center_right[i_row])
            if i_ap == 0:
                mask[i_row, :interp_stop] = True
            if i_ap == len(al) - 1:
                mask[i_row, interp_start:] = True
            bg[i_row, interp_start:interp_stop] = np.where(
                mask[i_row, interp_start:interp_stop],
                line_slope * np.arange(interp_start, interp_stop) + line_intersect,
                bg[i_row, interp_start:interp_stop],
            )

    print("Doing median filtering ...")
    bg1 = filters.median(bg, footprint=morphology.disk(median_sigma), mode="reflect")
    print("Doing gaussian filtering ...")
    bg2 = filters.gaussian(bg1, sigma=gaussian_sigma, mode="reflect")
    return bg2

# if __name__ == "__main__":
#     import matplotlib.pyplot as plt
#     fig, axs = plt.subplots(1, 2, figsize=(12, 6), sharex=True, sharey=True)
#     axs[0].imshow(np.log10(img))
#     axs[1].imshow(np.log10(bg1))
#     axs[0].plot(al.ap_center.T, np.arange(2048), "w-")
#     axs[1].plot(al.ap_center.T, np.arange(2048), "w-")
#
#     figure(figsize=(12, 4))
#     # plot(img[1000])
#     # plot(bg2[1000])
#     plot(img[0])
#     plot(bg[0])
#     plot(bg1[0])
#     plot(bg2[0])
#
#     figure()
#     plot(img[i_row])
#     plot(img[i_row] * mask[i_row])



import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import least_squares


def cost1d(p, x, y, pw=2):
    res = np.polynomial.chebyshev.chebval(x, p) - y
    return np.abs(res) ** (pw / 2)


def cost2d(p, x, y, z, deg=(1, 1), pw=2):
    res = np.polynomial.chebyshev.chebval2d(x, y, p.reshape(deg[0] + 1, deg[1] + 1)) - z
    return np.abs(res) ** (pw / 2)


class PolyFit1D:
    """
    1D polynomial fitter.

    Parameters
    ----------
    x : np.ndarray
        X data.
    y : np.ndarray
        Y data.
    deg : int
        The degree of polynomial in X dimension.
    pw : int
        The cost type is ``(abs(model-obs))**pw``.
    robust : bool
        If True, use median instead of mean to estimate initial parameters.
    """
    x_mean = 0
    x_scale = 0
    y_mean = 0
    y_scale = 0

    x_sc = 0
    y_sc = 0

    opt = None
    p = None
    deg = 0
    rms = 0

    def __init__(self, x, y, deg=1, pw=1, robust=True):
        # determine data scale
        if robust:
            self.x_scale = np.ptp(np.percentile(x, q=[84, 16])) / 2
            self.x_mean = np.percentile(x, q=50)
            self.y_scale = np.ptp(np.percentile(y, q=[84, 16])) / 2
            self.y_mean = np.percentile(y, q=50)
        else:
            self.x_scale = np.std(x)
            self.x_mean = np.mean(x)
            self.y_scale = np.std(y)
            self.y_mean = np.mean(y)

        # normalize data
        self.x_sc = (x - self.x_mean) / self.x_scale
        self.y_sc = (y - self.y_mean) / self.y_scale

        # fit data
        p0 = np.zeros((deg + 1,), dtype=float)
        self.opt = least_squares(cost1d, p0, method="lm", args=(self.x_sc, self.y_sc, pw))
        self.p = self.opt.x

        self.pw = pw
        self.robust = robust
        self.deg = deg

    def __call__(self, x):
        x_sc = (x - self.x_mean) / self.x_scale
        y_sc = np.polynomial.chebyshev.chebval(x_sc, self.p)
        return y_sc * self.y_scale + self.y_mean


class PolyFit2D:
    """
    2D polynomial fitter.

    Parameters
    ----------
    x : np.ndarray
        X data.
    y : np.ndarray
        Y data.
    z : np.ndarray
        Z data.
    deg : tuple
        The degrees of polynomials in X and Y dimensions.
    pw : int
        The cost type is ``(abs(model-obs))**pw``.
    robust : bool
        If True, use median instead of mean to estimate initial parameters.
    """
    x_mean = 0
    x_scale = 0
    y_mean = 0
    y_scale = 0
    z_mean = 0
    z_scale = 0

    x_sc = 0
    y_sc = 0
    z_sc = 0

    opt = None
    p = None
    deg = 0
    rms = 0

    def __init__(self, x, y, z, deg=(1, 2), pw=1, robust=True):
        # determine data scale
        if robust:
            self.x_scale = np.ptp(np.percentile(x, q=[84, 16])) / 2
            self.x_mean = np.percentile(x, q=50)
            self.y_scale = np.ptp(np.percentile(y, q=[84, 16])) / 2
            self.y_mean = np.percentile(y, q=50)
            self.z_scale = np.ptp(np.percentile(z, q=[84, 16])) / 2
            self.z_mean = np.percentile(z, q=50)
        else:
            self.x_scale = np.std(x)
            self.x_mean = np.mean(x)
            self.y_scale = np.std(y)
            self.y_mean = np.mean(y)
            self.z_scale = np.std(z)
            self.z_mean = np.mean(z)

        # normalize data
        self.x_sc = (x - self.x_mean) / self.x_scale
        self.y_sc = (y - self.y_mean) / self.y_scale
        self.z_sc = (z - self.z_mean) / self.z_scale

        # fit data
        p0 = np.zeros((deg[0] + 1, deg[1] + 1), dtype=float).flatten()
        self.opt = least_squares(cost2d, p0, method="lm", args=(self.x_sc, self.y_sc, self.z_sc, deg, pw))
        self.p = self.opt.x

        self.pw = pw
        self.robust = robust
        self.deg = deg

    def __call__(self, x, y):
        x_sc = (x - self.x_mean) / self.x_scale
        y_sc = (y - self.y_mean) / self.y_scale
        z_sc = np.polynomial.chebyshev.chebval2d(
            x_sc, y_sc, self.p.reshape(self.deg[0] + 1, self.deg[1] + 1))
        return z_sc * self.z_scale + self.z_mean


def test_poly1d():
    xx = np.linspace(1001, 1002, 100)
    yy = (xx - 1001.5) ** 2 * 100 + 100 + np.random.normal(0, 1, xx.shape)

    fig = plt.figure()
    plt.plot(xx, yy, 'x')
    poly1d1 = PolyFit1D(xx, yy, deg=2, pw=1)
    plt.plot(xx, poly1d1(xx), 'r-', label="deg=2, pw=1")
    poly1d2 = PolyFit1D(xx, yy, deg=2, pw=2)
    plt.plot(xx, poly1d2(xx), 'r-', label="deg=2, pw=2")
    poly1dnp = np.polyfit(xx, yy, deg=2)
    plt.plot(xx, np.polyval(poly1dnp, xx), color="gray", label="numpy.polyfit")
    plt.legend()
    plt.xlabel("X")
    plt.ylabel("Y")
    fig.tight_layout()


def test_poly2d():
    xx = np.random.uniform(1001, 1002, 1000)
    yy = np.random.uniform(2001, 2002, 1000)
    zz = (xx - 1001.5) ** 2 - (yy - 2001.5) ** 2 + 100

    xgrid, ygrid = np.meshgrid(
        np.linspace(1001, 1002, 100),
        np.linspace(2001, 2002, 100),
    )

    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    ax.scatter(xx, yy, zz, s=5, c="cyan", label="Data")
    poly2d = PolyFit2D(xx, yy, zz, deg=(2, 2), pw=2)

    zpred = poly2d(xgrid, ygrid)
    ax.plot_surface(
        xgrid, ygrid, zpred, rstride=1, cstride=1, cmap=plt.cm.coolwarm,
        linewidth=0, antialiased=False, alpha=.5, label="Poly2dFitter")
    # ax.legend()
    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")
    fig.tight_layout()


if __name__ == "__main__":
    test_poly1d()
    test_poly2d()



import glob


def glob_fits(root="/Users/cham/projects/song/star_spec"):
    """ Glob .fits files in a root directory and sort. """
    fps = glob.glob(root + "/**/*.fits", recursive=True)
    fps.sort()
    return fps


def test():
    fps = glob_fits()
    print(f"{len(fps)} files found!")
    print(fps)



import os
import yaml


PACKAGE_PATH = os.path.join(os.path.dirname(__file__))

with open(os.path.join(PACKAGE_PATH, "config/songcn_params.yml")) as f:
    SONGCN_PARAMS = yaml.safe_load(f)

with open(os.path.join(PACKAGE_PATH, "config/songcn_default.yml")) as f:
    SONGCN_DEFAULT = yaml.safe_load(f)


def test():
    print(SONGCN_DEFAULT)
    print(SONGCN_PARAMS)


if __name__ == "__main__":
    test()



import logging
from collections import namedtuple

import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from scipy.signal import argrelextrema

LocalMax = namedtuple(
    typename="LocalMax",
    field_names=[
        "num",  # number of maxima
        "max_ind",  # index of maxima
        "max_val",  # value of maxima
        "max_val_interp",  # value interpolated from neighboring minima
        "max_snr",  # SNR of maxima (max/max_val_interp)
        "min_ind",  # index of minima
        "side_ind",  # index of two sides
        "side_mask",  # True if out of bounds
    ]
)


def generate_pulse_kernel(kernel_width=15):
    """ Generate a pulse kernel. """
    kernel = np.zeros((kernel_width + 2), dtype=float)
    kernel[1:-1] = 1
    return kernel


def find_local_max(arr, kernel_width=15):
    """ Find local max in 1D array.

    Parameters
    ----------
    arr : np.array
        The 1D array.
    kernel_width : int
        The kernel width in number of pixels.

    """
    arr_convolved = np.convolve(arr, generate_pulse_kernel(kernel_width=kernel_width), mode="same")
    # find local max using scipy.signal.argrelextrema
    ind_local_max = argrelextrema(arr_convolved, np.greater_equal, order=kernel_width, mode='clip')[0]
    logging.info(f"{len(ind_local_max)} maxima found")
    # interpolate for local min
    ind_local_max_delta = np.diff(ind_local_max) / 2
    ind_local_min_derived = np.hstack(
        (
            ind_local_max[:-1] - ind_local_max_delta,
            ind_local_max[-1] - ind_local_max_delta[-1],
            ind_local_max[-1] + ind_local_max_delta[-1],
        )
    ).astype(int)
    # calculate SNR for local max
    ind_two_sides = np.array([
        ind_local_min_derived[:-1],
        ind_local_min_derived[1:]
    ])
    ind_two_sides_mask = np.logical_or(ind_two_sides < 0, ind_two_sides > len(arr) - 1)
    ind_two_sides_valid = np.where(ind_two_sides_mask, 0, ind_two_sides)  # do not go out of bounds
    # estimate SNR of local max, clip out-of-bounds values
    interp_val_local_max = np.ma.MaskedArray(
        data=arr_convolved[ind_two_sides_valid],
        mask=ind_two_sides_mask,
    ).mean(axis=0)
    assert interp_val_local_max.mask.sum() == 0
    interp_val_local_max = interp_val_local_max.data
    val_local_max = arr_convolved[ind_local_max]
    snr_local_max = val_local_max / interp_val_local_max
    return LocalMax(
        num=len(ind_local_max),
        max_ind=ind_local_max,
        max_val=val_local_max,
        max_val_interp=interp_val_local_max,
        max_snr=snr_local_max,
        min_ind=ind_local_min_derived,
        side_ind=ind_two_sides,
        side_mask=ind_two_sides_mask,
    )


def find_local_max_2d(img, pos_init=(1, 1), extra_bin=1, kernel_width=10):
    """ slice by slice """
    is_local_max = np.zeros_like(img, dtype=bool)
    for i_row in range(img.shape[0]):
        starting_row = i_row - extra_bin
        stop_row = i_row + 1 + extra_bin
        this_slice = img[starting_row:stop_row].sum(axis=0)
        localmax = find_local_max(this_slice, kernel_width=kernel_width)
        is_local_max[i_row, localmax.max_ind] = True
        print(f"irow = {i_row}, {localmax.num} max found")
    return is_local_max


def trace_one_aperture(img, init_pos=(1000, 993), extra_bin=1, kernel_width=10, maxdev=5):
    """
    Given an initial position, search for the aperture.

    The image will be convolved with a pulse kernel slice by slice.
    Then the search starts from initial position row by row.

    Parameters
    ----------
    img : np.ndarray
        The 2D image.
    init_pos : tuple
        A tuple of (row, col) of the initial position.
    extra_bin : int
        The number of extra binning rows to enhance SNR.
    kernel_width : int
        The width of the pulse kernel.
    maxdev : int
        The max deviation of ap_col allowed in consecutive two rows.

    Returns
    -------
    tuple
        (ap_col, ap_mask) where ap_col is the aperture center index and ap_mask the validity.
        If the ap_col is bad, ap_mask is set True.
    """
    n_row, n_col = img.shape
    init_row, init_col = init_pos
    if kernel_width is not None:
        # convolve image slice by slice (row by row)
        img = np.array(
            [
                np.convolve(
                    img[i_row - extra_bin - extra_bin:i_row + 1 + extra_bin].sum(axis=0),
                    generate_pulse_kernel(kernel_width=kernel_width),
                    mode="same"
                ) for i_row in np.arange(n_row)
            ]
        )

    ap_col = np.zeros(n_row, dtype=int)
    ap_mask = np.zeros(n_row, dtype=bool)
    ap_col[init_row] = init_col
    # search in current row
    i_row = init_row  # this is special
    chunk_data = img[i_row, ap_col[i_row] - 2 * kernel_width:ap_col[i_row] + 1 + 2 * kernel_width]
    argmax = argrelextrema(chunk_data, np.greater_equal, order=kernel_width, mode='clip')[0]
    ind_mindev = np.argmin(np.abs(argmax - 2 * kernel_width))
    if np.min(np.abs(argmax[ind_mindev] - 2 * kernel_width)) <= maxdev:
        ap_col[i_row] = ap_col[i_row] - 2 * kernel_width + argmax[ind_mindev]
    else:
        logging.warning(f"Warning: deviation({np.min(np.abs(argmax[ind_mindev] - 2 * kernel_width)):d})"
                        f" larger than maxdev({maxdev}) in Row {i_row}")
        ap_col[i_row] = ap_col[i_row]
        ap_mask[i_row] = True

    # search the above
    for i_row in np.arange(init_row)[::-1]:
        # try:
        chunk_start = max(0, ap_col[i_row + 1] - 2 * kernel_width)
        chunk_stop = min(n_col - 1, ap_col[i_row + 1] + 1 + 2 * kernel_width)
        chunk_data = img[i_row, chunk_start:chunk_stop]
        argmax = argrelextrema(chunk_data, np.greater_equal, order=kernel_width, mode='clip')[0]
        ind_mindev = np.argmin(np.abs(argmax + chunk_start - ap_col[i_row + 1]))
        if np.min(np.abs(argmax[ind_mindev] - 2 * kernel_width)) <= maxdev:
            ap_col[i_row] = ap_col[i_row + 1] - 2 * kernel_width + argmax[ind_mindev]
        else:
            logging.warning(f"Warning: deviation({np.min(np.abs(argmax[ind_mindev] - 2 * kernel_width)):d})"
                            f" larger than maxdev({maxdev}) in Row {i_row}")
            ap_col[i_row] = ap_col[i_row + 1]
            ap_mask[i_row] = True
        # except:
        #     ap_col[i_row] = ap_col[i_row + 1]
        #     ap_mask[i_row] = True

    # search the below
    for i_row in np.arange(init_row + 1, n_row):
        # try:
        chunk_start = max(0, ap_col[i_row - 1] - 2 * kernel_width)
        chunk_stop = min(n_col - 1, ap_col[i_row - 1] + 1 + 2 * kernel_width)
        chunk_data = img[i_row, chunk_start:chunk_stop]
        argmax = argrelextrema(chunk_data, np.greater_equal, order=kernel_width, mode='clip')[0]
        ind_mindev = np.argmin(np.abs(argmax + chunk_start - ap_col[i_row - 1]))
        if np.min(np.abs(argmax[ind_mindev] - 2 * kernel_width)) <= maxdev:
            ap_col[i_row] = ap_col[i_row - 1] - 2 * kernel_width + argmax[ind_mindev]
        else:
            logging.warning(f"Warning: deviation({np.min(np.abs(argmax[ind_mindev] - 2 * kernel_width)):d})"
                            f" larger than maxdev({maxdev}) in Row {i_row}")
            ap_col[i_row] = ap_col[i_row - 1]
            ap_mask[i_row] = True
        # except:
        #     ap_col[i_row] = ap_col[i_row - 1]
        #     ap_mask[i_row] = True

    return ap_col, ap_mask


if __name__ == "__main__":
    fp = "/Users/cham/projects/song/star_spec/20191105/night/ext/masterflat_20191105_slit5.fits"
    img = fits.getdata(fp)
    central_ind = int(img.shape[0] / 2)
    central_slice = img[central_ind:central_ind + 10].sum(axis=0)
    for kernel_width in np.arange(5, 30):
        localmax = find_local_max(arr=central_slice, kernel_width=kernel_width)
        print(
            f"kernel_width: {kernel_width}, number of max: {localmax.num}, "
            f"median SNR: {np.median(localmax.max_snr):.2f}, mean SNR: {np.mean(localmax.max_snr):.2f}"
        )

    localmax = find_local_max(arr=central_slice, kernel_width=15)

    fig, ax = plt.subplots(1, 1, figsize=(8, 8))
    ax.imshow(np.log10(img), extent=(-.5, 2047.5, -.5, 2047.5,), origin="lower")
    for i_ap in np.arange(localmax.num):
        ap_col, ap_mask = trace_one_aperture(
            img, init_pos=(central_ind, localmax.max_ind[i_ap]), extra_bin=1, kernel_width=15, maxdev=3)
        ax.plot(ap_col, np.arange(img.shape[0]))
        print(f"i_ap = {i_ap}, masksum = {sum(ap_mask)}")
    plt.show()



import datetime

import joblib
import matplotlib.pyplot as plt
import numpy as np
from scipy.linalg import toeplitz


class PV2002:
    """
    P&V (2002) solver.

    Parameters
    ----------
    S_lambda_x : np.ndarray
        (n_lambda, n_width), the spectral image.
    remainders : np.ndarray
        (n_lambda,), the aperture center remainders.
    osr : int
        The over sampling rate.
    Lrel : float
        The relative regularization parameter. The default value is 1.
    zero_wing : int
        The number of edge pixels which are set to zero. The default value is 1.
    clip_sigma : float
        Clip the rows with deviations larger than {clip_sigma} * median deviation.
    xx : np.ndarray
        The X coordinates.
    yy : np.ndarray
        The Y coordinates.

    Attributes
    ----------
    S_lambda_x : np.ndarray
        (n_lambda, n_width), the spectral image.
    f_lambda_sum : np.ndarray
        (n_lambda,), the sum-extracted 1D spectrum.
    f_lambda_0 : np.ndarray
        (n_lambda,), the optimally extracted 1D spectrum in the last iteration.
    f_lambda_1 : np.ndarray
        (n_lambda,), the optimally extracted 1D spectrum in the current iteration.
    g_j_0 : np.ndarray
        ((n_width+1) * osr,), the optimally extracted spatial profile in the last iteration.
    g_j_1 : np.ndarray
        ((n_width+1) * osr,), the optimally extracted spatial profile in the current iteration.

    """

    def __init__(self, S_lambda_x, remainders, osr=10, Lrel=1, zero_wing=1, clip_sigma=3, xx=None, yy=None):
        # important attributes
        self.xx = xx
        self.yy = yy
        self.S_lambda_x = S_lambda_x
        self.n_lambda, self.n_width = self.S_lambda_x.shape
        self.f_lambda_sum = np.einsum("ij->i", S_lambda_x) - remainders * S_lambda_x[:, 0] - (
                1 - remainders) * S_lambda_x[:, -1]
        self.l_lambda_0 = self.l_lambda_1 = None  # sky light spectrum, initialized with None
        self.f_lambda_0 = self.f_lambda_1 = None  # 1D spectrum, initialized with None
        self.g_j_0 = self.g_j_1 = None  # spatial profile, initialized with None
        self.remainders = remainders
        self.osr = osr
        self.L = np.median(self.f_lambda_sum) ** 2 * Lrel
        self.zero_wing = zero_wing
        # derived quantities
        self.j_max = osr * (self.n_width + 1)
        self.B_j_k = self.eval_B_j_k(j_max=self.j_max)
        self.w_lambda_x_j = self.eval_w_lambda_x_j(remainders, osr=osr, n_pixel=self.n_width)
        self.A_j_k = None
        self.R_k = None
        self.C_lambda = None
        self.D_lambda = None
        self.S_rec = None
        # pixel mask
        self.lambda_good = np.ones(self.n_lambda, dtype=bool)
        # iteration info
        self.n_iter = 0
        self.f_lambda_history = []
        self.g_j_history = []
        self.dt = 0.
        self.d_g_j = 0.
        self.d_f_lambda = 0.
        self.d_l_lambda = 0.
        self.maxdev_S = 0.
        self.meddev_S = 0.
        self.clip_sigma = clip_sigma

    def reduce(self, n_iter=10, tol_f_lambda=1e-3, tol_g_j=1e-10):
        """
        Reduce spectrum image iteratively until reaching tolerances.

        Parameters
        ----------
        n_iter: int
            The max iteration number.
        tol_f_lambda : float
            The tolerance of 1D spectrum.
        tol_g_j : float
            The tolerance of profile (slit function).
        """
        for i_iter in range(n_iter):
            self.iterate()
            if np.linalg.norm(self.f_lambda_1 - self.f_lambda_0, np.inf) < tol_f_lambda \
                    and np.linalg.norm(self.g_j_1 - self.g_j_0, np.inf) < tol_g_j:
                break

    def iterate(self):
        """ Do one iteration with P&V 2002 algorithm. """
        t_0 = datetime.datetime.now()

        # prepare to iterate
        self.f_lambda_0 = self.f_lambda_1
        self.g_j_0 = self.g_j_1

        if self.f_lambda_0 is None:
            self.f_lambda_0 = self.f_lambda_sum
            self.f_lambda_history.append(self.f_lambda_0)
            self.g_j_0 = np.ones((self.n_width + 1) * self.osr, dtype=float) / (self.n_width + 1)

        # solve new spatial profile
        self.A_j_k = self.eval_A_j_k(self.f_lambda_0[self.lambda_good], self.w_lambda_x_j[self.lambda_good])
        self.R_k = self.eval_R_k(
            self.S_lambda_x[self.lambda_good], self.f_lambda_0[self.lambda_good], self.w_lambda_x_j[self.lambda_good]
        )
        self.g_j_1 = np.linalg.solve(self.A_j_k + self.L * self.B_j_k, self.R_k)
        if self.zero_wing:
            self.g_j_1[:self.osr * self.zero_wing] = 0
            self.g_j_1[-self.osr * self.zero_wing:] = 0
        self.g_j_1[self.g_j_1 < 0] = 0
        self.g_j_1 *= self.osr / self.g_j_1.sum()
        # solve new spectrum with new spatial profile
        self.C_lambda = self.eval_C_lambda(self.S_lambda_x, self.w_lambda_x_j, self.g_j_1)
        self.D_lambda = self.eval_D_lambda(self.w_lambda_x_j, self.g_j_1)
        self.f_lambda_1 = self.C_lambda / self.D_lambda
        # reconstruct image
        self.S_rec = np.einsum("i,ijk,k->ij", self.f_lambda_1, self.w_lambda_x_j, self.g_j_1)
        # record iteration number
        self.n_iter += 1
        # append results in history
        self.f_lambda_history.append(self.f_lambda_1)
        self.g_j_history.append(self.g_j_1)
        # verbose info
        self.dt = datetime.datetime.now() - t_0
        # evaluate variations in spectrum and profile
        self.d_g_j = np.linalg.norm(self.g_j_1 - self.g_j_0, np.inf)
        self.d_f_lambda = np.linalg.norm(self.f_lambda_1 - self.f_lambda_0, np.inf)
        # evaluate deviation from input images
        self.maxdev_S = np.linalg.norm(self.S_lambda_x - self.S_rec, np.inf)
        self.meddev_S = np.median(np.max(np.abs(self.S_lambda_x - self.S_rec), axis=1))

        self.lambda_good &= np.all(np.abs(self.S_lambda_x - self.S_rec) < self.clip_sigma * self.meddev_S, axis=1)

        print(f"Finish {self.n_iter}th iteration: D(t)={self.dt.total_seconds():.2f} sec! \n"
              f"    - D(g_j)      = {self.d_g_j:.2e},\n"
              f"    - D(f_lambda) = {self.d_f_lambda:.2e},\n"
              f"    - MedDev(S)   = {self.meddev_S:.2e},\n"
              f"    - MaxDev(S)   = {self.maxdev_S:.2e},\n"
              f"    - N_clip      = {np.sum(~self.lambda_good)} / {self.n_lambda}")
        return

    def iterate_ipv(self, sky_regularization=1.):
        """ Do one iteration with P&V 2002 algorithm. """
        t_0 = datetime.datetime.now()

        # prepare to iterate
        self.f_lambda_0 = self.f_lambda_1
        self.l_lambda_0 = self.l_lambda_1
        self.g_j_0 = self.g_j_1

        if self.f_lambda_0 is None:
            self.f_lambda_0 = self.f_lambda_sum
            self.f_lambda_history.append(self.f_lambda_0)
            self.g_j_0 = np.ones((self.n_width + 1) * self.osr, dtype=float) / (self.n_width + 1)
            self.l_lambda_0 = np.zeros_like(self.f_lambda_0, dtype=float)

        # solve new spatial profile
        self.A_j_k = self.eval_A_j_k(self.f_lambda_0[self.lambda_good], self.w_lambda_x_j[self.lambda_good])
        self.R_k = self.eval_R_k(
            self.S_lambda_x[self.lambda_good] - self.l_lambda_0[self.lambda_good, None],
            self.f_lambda_0[self.lambda_good],
            self.w_lambda_x_j[self.lambda_good]
        )
        self.g_j_1 = np.linalg.solve(self.A_j_k + self.L * self.B_j_k, self.R_k)
        if self.zero_wing:
            self.g_j_1[:self.osr * self.zero_wing] = 0
            self.g_j_1[-self.osr * self.zero_wing:] = 0
        self.g_j_1[self.g_j_1 < 0] = 0
        self.g_j_1 *= self.osr / self.g_j_1.sum()
        # solve new spectrum with new spatial profile
        self.C_lambda = self.eval_C_lambda(self.S_lambda_x - self.l_lambda_0[:, None], self.w_lambda_x_j, self.g_j_1)
        self.D_lambda = self.eval_D_lambda(self.w_lambda_x_j, self.g_j_1)
        self.f_lambda_1 = self.C_lambda / self.D_lambda
        # reconstruct image
        self.S_rec = np.einsum("i,ijk,k->ij", self.f_lambda_1, self.w_lambda_x_j, self.g_j_1)
        # solve skylight
        self.l_lambda_1 = np.mean(self.S_lambda_x - self.S_rec, axis=1) - sky_regularization
        # force skylight to be positive
        self.l_lambda_1[self.l_lambda_1 < 0] = 0.
        # record iteration number
        self.n_iter += 1
        # append results in history
        self.f_lambda_history.append(self.f_lambda_1)
        self.g_j_history.append(self.g_j_1)
        # verbose info
        self.dt = datetime.datetime.now() - t_0
        # evaluate variations in spectrum and profile
        self.d_g_j = np.linalg.norm(self.g_j_1 - self.g_j_0, np.inf)
        self.d_f_lambda = np.linalg.norm(self.f_lambda_1 - self.f_lambda_0, np.inf)
        self.d_l_lambda = np.linalg.norm(self.l_lambda_1 - self.l_lambda_0, np.inf)
        # evaluate deviation from input images
        self.maxdev_S = np.linalg.norm(self.S_lambda_x - self.S_rec, np.inf)
        self.meddev_S = np.median(np.max(np.abs(self.S_lambda_x - self.S_rec), axis=1))

        self.lambda_good &= np.all(np.abs(self.S_lambda_x - self.S_rec) < self.clip_sigma * self.meddev_S, axis=1)

        print(
            f"Finish {self.n_iter}th iteration: D(t)={self.dt.total_seconds():.2f} sec! \n"
            f"    - D(g_j)      = {self.d_g_j:.2e},\n"
            f"    - D(f_lambda) = {self.d_f_lambda:.2e},\n"
            f"    - D(l_lambda) = {self.d_l_lambda:.2e},\n"
            f"    - MedDev(S)   = {self.meddev_S:.2e},\n"
            f"    - MaxDev(S)   = {self.maxdev_S:.2e},\n"
            f"    - N_clip      = {np.sum(~self.lambda_good)} / {self.n_lambda}"
        )
        return

    @staticmethod
    def eval_B_j_k_(j_max=320):
        """ Evaluate B_j_k, the tri-diagonal matrix with toeplitz method. (Deprecated) """
        c = np.zeros(j_max, dtype=float)
        r = np.zeros(j_max, dtype=float)
        c[0] = 2
        c[1] = -1
        r[1] = -1
        B_j_k = toeplitz(c, r)
        B_j_k[0, 0] = 1
        B_j_k[-1, -1] = 1
        return B_j_k

    @staticmethod
    def eval_B_j_k(j_max=320):
        """
        Evaluate B_j_k, the tri-diagonal matrix.

        Parameters
        ----------
        j_max : int
            The max j value, which is `(n_width + 1 * osr)`.

        Returns
        -------
        np.ndarray
            The B_j_k.
        """
        B_j_k = np.diag(np.ones(j_max) * 2., k=0) + \
                np.diag(np.ones(j_max - 1) * -1., k=1) + \
                np.diag(np.ones(j_max - 1) * -1., k=-1)
        B_j_k[0, 0] = 1
        B_j_k[-1, -1] = 1
        return B_j_k

    @staticmethod
    def eval_A_j_k(f_lambda, w_lambda_x_j):
        """
        Evaluate A_j_k.

        Parameters
        ----------
        f_lambda : ndarray
            (n_lambda,)
        w_lambda_x_j : ndarray
            (n_lambda, n_width, (n_width+1)*osr)

        Returns
        -------
        ndarray
            ((n_width+1)*osr, (n_width+1)*osr)
        """
        # return np.einsum("i,ijkh->kh", f_lambda**2, np.einsum("...i,...j->...ij", w_lambda_x_j, w_lambda_x_j))
        return np.einsum("i,ijk,ijh->kh", np.square(f_lambda), w_lambda_x_j, w_lambda_x_j)

    @staticmethod
    def eval_R_k(S_lambda_x, f_lambda, w_lambda_x_j):
        """
        Evaluate R_k.

        Parameters
        ----------
        S_lambda_x : ndarray
            (n_lambda, n_width)
        f_lambda : ndarray
            (n_lambda,)
        w_lambda_x_j : ndarray
            (n_lambda, n_width, (n_width+1)*osr)

        Returns
        -------
        ndarray
            ((n_width+1)*osr, (n_width+1)*osr)
        """
        return np.einsum("ij,i,ijk->k", S_lambda_x, f_lambda, w_lambda_x_j)

    @staticmethod
    def eval_C_lambda(S_lambda_x, w_lambda_x_j, g_j):
        """
        Evaluate C_lambda.

        Parameters
        ----------
        S_lambda_x : ndarray
            (n_lambda, n_width)
        w_lambda_x_j : ndarray
            (n_lambda, n_width, (n_width+1)*osr)
        g_j : ndarray
            ((n_width+1)*osr,)

        Returns
        -------
        ndarray
            ((n_width+1)*osr, (n_width+1)*osr)
        """
        return np.einsum("ij,ijk,k->i", S_lambda_x, w_lambda_x_j, g_j)

    @staticmethod
    def eval_D_lambda(w_lambda_x_j, g_j):
        """
        Evaluate D_lambda.

        Parameters
        ----------
        w_lambda_x_j : ndarray
            (n_lambda, n_width, (n_width+1)*osr)
        g_j : ndarray
            ((n_width+1)*osr,)

        Returns
        -------
        ndarray
            (n_lambda,)
        """
        return np.einsum("ij->i", np.square(np.einsum("ijk,k->ij", w_lambda_x_j, g_j)))

    @staticmethod
    def eval_w_lambda(remainder=0.0247, osr=10, n_pixel=31):
        """ Evaluate w_lambda_x_j at a given lambda. """
        # evaluate start index of period
        start_ind = int(np.floor(remainder * osr))
        # construct period
        period = np.ones((osr + 1,), dtype=float) / osr
        period[-1] = np.mod(remainder, 1. / osr)
        period[0] = 1. / osr - period[-1]
        # construct w_lambda
        w_lambda = np.zeros((n_pixel, (n_pixel + 1) * osr), dtype=float)
        for i_pixel in range(n_pixel):
            w_lambda[i_pixel, start_ind + i_pixel * osr:start_ind + (i_pixel + 1) * osr + 1] = period
        return w_lambda[::-1]

    @staticmethod
    def eval_w_lambda_x_j(remainders, osr=10, n_pixel=31):
        """
        Evaluate w_lambda_x_j from aperture center remainders, over-sampling rate, and aperture width.

        Parameters
        ----------
        remainders : array-like
            The aperture center remainder.
        osr : int, optional
            The over-sampling rate. The default is 10.
        n_pixel : int, optional
            The aperture width (pixels). The default is 31.

        Returns
        -------
        np.ndarray
            W_lambda_x.

        """
        return np.array(
            [
                PV2002.eval_w_lambda(remainder=remainder, osr=osr, n_pixel=n_pixel) for remainder in remainders
            ]
        )

    def plot_result(self, profile_shift=0.002):
        assert self.xx is not None and self.yy is not None

        fig = plt.figure(figsize=(10, 12))

        ax0 = fig.add_subplot(3, 2, 1, title="Input image")
        I = ax0.imshow(self.S_lambda_x, aspect="auto", interpolation="nearest", cmap=plt.cm.jet,
                       vmin=np.min(self.S_rec),
                       vmax=np.max(self.S_rec))
        plt.colorbar(I, ax=ax0)
        ax0.set_xlabel("Rectified $X$ [pixel]")
        ax0.set_ylabel("$Y$ [pixel]")

        ax1 = fig.add_subplot(3, 2, 2, title="Reconstructed image")
        I = ax1.imshow(
            self.S_rec, aspect="auto", interpolation="nearest", cmap=plt.cm.jet, vmin=np.min(self.S_rec),
            vmax=np.max(self.S_rec)
        )
        plt.colorbar(I, ax=ax1)
        ax1.set_xlabel("Rectified $X$ [pixel]")
        ax1.set_ylabel("$Y$ [pixel]")

        ax0 = fig.add_subplot(3, 2, 3, projection='3d', title="Input image")
        ax1 = fig.add_subplot(3, 2, 4, projection='3d', title="Reconstructed image")
        ax0.plot_surface(
            self.xx, self.yy, self.S_lambda_x, cmap=plt.cm.jet, vmin=np.min(self.S_rec), vmax=np.max(self.S_rec)
        )
        ax1.plot_surface(
            self.xx, self.yy, self.S_rec, cmap=plt.cm.jet, vmin=np.min(self.S_rec), vmax=np.max(self.S_rec),
        )
        ax0.set_ylabel("$X$ (spatial direction) [pixel]")
        ax0.set_xlabel("$Y$ (dispersion direction) [pixel]")
        ax1.set_ylabel("$X$ (spatial direction) [pixel]")
        ax1.set_xlabel("$Y$ (dispersion direction) [pixel]")
        ax0.set_zlim(ax1.get_zlim())
        ax0.set_zlabel("Counts [ADU]")
        ax1.set_zlabel("Counts [ADU]")
        ax0.view_init(30, -10, 0)
        ax1.view_init(30, -10, 0)

        # plot profile
        ax_g = fig.add_subplot(3, 2, 5, title="Spatial profile")
        # data
        ax_g.plot(
            np.arange(self.n_width).reshape(-1, 1) + 1 - self.remainders,
            self.S_lambda_x.T / self.f_lambda_sum,
            color="gray", zorder=-1
        )
        # estimated profile
        for i, g in enumerate(self.g_j_history):
            ax_g.plot(np.arange(len(g)) / self.osr - .5 + .5 / self.osr,  # this is tricky
                      g[::-1] + (self.n_iter - 1 - i) * profile_shift)
            print("sum(g): ", g.sum())
        ax_g.set_xlabel("Corrected $X$ [pixel]")
        ax_g.set_ylabel("Profile density $\\times$ Over sampling rate")

        # plot spectrum
        ax_f = fig.add_subplot(3, 2, 6, title="Extracted spectrum")
        for i, f in enumerate(self.f_lambda_history):
            if i == 0:
                plt.plot(f, label=f"Iter {i}", color="k")
            else:
                plt.plot(f, label=f"Iter {i}")
            print("median(f): ", np.median(f))
        ax_f.legend()
        ax_f.set_xlabel("$\\lambda$ [pixel]")
        ax_f.set_ylabel("Counts [ADU]")

        fig.tight_layout()
        plt.show()
        # fig.subplots_adjust(top=0.95)
        return fig


def test_downsample_spectral_profile():
    # %timeit eval_w_lambda(remainder=0.247, osr=10, n_pixel=31)
    # %timeit eval_w(remainders=ap.ap_center_remainder[:256], osr=10, n_pixel=31).shape
    n_pixel = 31
    osr = 10
    test_g = np.ones((n_pixel + 1) * osr).reshape(-1, 1)
    test_w = PV2002.eval_w_lambda_x_j(remainders=np.random.uniform(0, 1, size=256), osr=10, n_pixel=31)
    print(np.linalg.norm(test_w @ test_g - 1))
    assert np.linalg.norm(test_w @ test_g - 1) < 1e-10
    # (256, 31, 320) @ (320,) = (256, 31)
    return


def generate_gaussian_profile(xx, center=0, width=1., amplitude=1.):
    yy = np.exp(-0.5 * ((xx - center) / width) ** 2) * amplitude
    return yy


def test_pv2002():
    # load data
    data = joblib.load("/Users/cham/PycharmProjects/songcn/songcn/data/test_pv2002/spec2d.z")
    ap_xx = data['ap_xx']
    ap_yy = data['ap_yy']
    ap_img = data['ap_img']
    ap_remainders = data['ap_remainders']
    # ap_img[10, 5:8] = 50000
    pv = PV2002(S_lambda_x=ap_img, remainders=ap_remainders, osr=5, Lrel=10, xx=ap_xx, yy=ap_yy)

    pv.reduce(n_iter=10, tol_f_lambda=1e-3, tol_g_j=1e-10)
    fig = pv.plot_result()
    return fig


if __name__ == "__main__":
    test_downsample_spectral_profile()
    fig = test_pv2002()



"""
Make sure the image has a vertical spatial axis and a horizontal dispersion axis.
"""

import numpy as np


def trace_local_max(img, starting_row=None, starting_col=None, maxdev=10, fov=20, ntol=5):
    """ trace aperture from a starting point (row, col) """
    nrow, ncol = img.shape
    assert ntol < starting_col < ncol - ntol
    # find the local max for this column
    this_ind_row = -np.ones(nrow, dtype=int)
    # this_ind_col = np.arange(ncol, dtype=int)
    # search this column
    this_ind_row[starting_col] = find_local_max(
        img[:, starting_col], starting_row, maxdev=maxdev, fov=fov)
    # search left
    for icol in range(starting_col - 1, -1, -1):
        # find the closest finite value
        icol_right = np.min((icol + 1 + ntol, starting_col + 1))
        icol_closest_valid = np.where(this_ind_row[icol + 1:icol_right] > 0)[0]
        if len(icol_closest_valid) > 0:
            this_ind_row[icol] = find_local_max(
                img[:, icol], this_ind_row[icol + 1:icol_right][icol_closest_valid[0]], maxdev=maxdev, fov=fov)
        else:
            break
    # search right
    for icol in range(starting_col + 1, ncol, 1):
        # find the closest finite value
        icol_left = np.max((icol - ntol, starting_col))
        icol_closest_valid = np.where(this_ind_row[icol_left:icol] > 0)[0]
        if len(icol_closest_valid) > 0:
            this_ind_row[icol] = find_local_max(
                img[:, icol], this_ind_row[icol_left:icol][icol_closest_valid[-1]], maxdev=maxdev, fov=fov)
        else:
            break
    return this_ind_row


def find_local_max(img, irow=None, maxdev=10, fov=20):
    coldata = img[irow]
    if irow < 0:
        return -1
    indmax = np.argmax(coldata[irow - fov:irow + fov + 1])
    if np.abs(indmax - fov) > maxdev:
        return -1
    else:
        return indmax + irow - fov



from songcn import SongCalendar

sc = SongCalendar.from_dir(root="/Users/cham/projects/song/star_spec")
sc.pprint_all()

sn = sc.get_one_night("20191029")
print(f"SongNight contains {len(sc)} files")
sn.pprint_all()
print("Selected BIAS:", sn.select({"IMAGETYP": "BIAS"}, method="random"))
# sn.master_bias()
# sn.master_flat()


#%%

from song.song import Song
song = Song.init(
    rootdir="/Users/cham/projects/song/star_spec",
    date="20191030",
    jdnight=True,
    n_jobs=-1,
    verbose=True,
    subdir="night",
    node=2
)




import numpy as np
from laspec.extern.interpolate import SmoothSpline
from scipy.interpolate import interp1d
from scipy.stats import binned_statistic
import joblib


####################################
# aperture processing functions
####################################

def get_aperture_section(im, ap_center_interp, ap_width=15):
    """ get aperture section and coordinates
    
    Parameters
    ----------
    im : ndarray
        target image.
    ap_center_interp : ndarray (npix,)
        ap_center_interp.
    ap_width : float, optional
        ap_width. The default is 15.

    Returns
    -------
    ap_im : ndarray
        DESCRIPTION.
    ap_im_xx : ndarray
        x coordinates.
    ap_im_yy : ndarray
        y coordinates.
    ap_im_xx_cor : ndarray
        x offset from center.

    """
    n_rows, n_cols = im.shape
    im_xx, im_yy = np.meshgrid(np.arange(n_cols), np.arange(n_rows))

    # cut image to get this aperture region
    ap_center_interp_floor = np.floor(ap_center_interp).astype(int)
    # ap_center_interp_rmdr= ap_center_interp-ap_center_interp_floor

    # ap_im_nrows = n_rows  # not used
    ap_im_ncols = ap_width*2+2

    ap_im_xx = ap_center_interp_floor.reshape(-1, 1)+np.arange(-ap_width, ap_width+2)
    ap_im_yy = im_yy[:, :ap_im_ncols]

    ap_im_xx_flat = ap_im_xx.flatten()
    ap_im_yy_flat = ap_im_yy.flatten()
    ap_im_flat = np.zeros_like(ap_im_xx_flat, dtype=np.float)

    ind_valid = (ap_im_xx_flat >= 0) & (ap_im_xx_flat <= n_cols-1)
    ap_im_flat[ind_valid] = im[ap_im_yy_flat[ind_valid], ap_im_xx_flat[ind_valid]]
    ap_im = ap_im_flat.reshape(ap_im_xx.shape)

    ap_im_xx_cor = ap_im_xx - ap_center_interp.reshape(-1, 1)
    return ap_im, ap_im_xx, ap_im_yy, ap_im_xx_cor


def extract_profile_simple(ap_im, ap_im_xx_cor, ap_width, profile_oversample=10, profile_smoothness=1e-1):

    # 1. normalize image along x
    ap_im_norm = ap_im / np.sum(ap_im, axis=1).reshape(-1, 1)
    # 2. extract profile (quite good so far) for each chunk
    prof_x = np.arange(-ap_width - 1, ap_width + 1, 1 / profile_oversample)
    sp_prof_init = SmoothSpline(
        ap_im_xx_cor.flatten(),
        ap_im_norm.flatten(),
        p=profile_smoothness)
    prof_init = sp_prof_init(prof_x)
    prof_init[prof_init < 0] = 0

    return prof_x, prof_init


def extract_profile(ap_im, ap_im_xx_cor, profile_smoothness=1e-2, n_chunks=8,
                    ap_width=15., profile_oversample=10., ndev=4):
    """ extract profile from aperture image ap_im 
    
    Parameters
    ----------
    ap_im : ndarray
        The target image.
    ap_im_xx_cor : ndarray
        x offset.
    profile_smoothness : float, optional
        The profile smoothness, passed to SmoothSpline. The default is 1e-2.
    n_chunks : int, optional
        The number of chunks. The default is 8.
    ap_width : float, optional
        The aperture width. The default is 15..
    profile_oversample : int, optional
        The oversampling factor of the profile. The default is 10.
    ndev : float, optional
        The number of sigmas when clipping outliers. The default is 4.

    Returns
    -------
    profs_rebin : ndarray
        The normalized rebinned profile.

    """
    # 0. determine chunk length
    n_rows = ap_im.shape[0]
    chunk_len = np.int(n_rows / n_chunks)
    chunk_centers = chunk_len * (np.arange(n_chunks) + 0.5)
    chunk_start = chunk_len * np.arange(n_chunks)

    # 1. normalize image along x
    ap_im_norm = ap_im / np.sum(ap_im, axis=1).reshape(-1, 1)

    # 2. extract profile (quite good so far) for each chunk
    prof_out = np.zeros_like(ap_im_norm, dtype=bool)
    prof_x = np.arange(-ap_width - 1, ap_width + 1 + 1 / profile_oversample, 1 / profile_oversample)
    profs = []
    prof_xnode = np.arange(-ap_width - 1, ap_width + 1 + 1)
    nnode = len(prof_xnode)
    for istart in chunk_start:
        # first smooth and calculate residuals
        this_prof_res = SmoothSpline(
            ap_im_xx_cor[istart:istart + chunk_len].flatten(),
            ap_im_norm[istart:istart + chunk_len].flatten(),
            p=profile_smoothness)(ap_im_xx_cor[istart:istart + chunk_len]) - ap_im_norm[istart:istart + chunk_len]
        # outlier index
        this_prof_out = np.zeros_like(ap_im_norm[istart:istart + chunk_len], dtype=bool)
        # clipping each node
        for inode, this_prof_xnode in enumerate(prof_xnode):
            if inode == 0 or inode == nnode:
                this_prof_xnode_ind = np.abs(ap_im_xx_cor[istart:istart + chunk_len] - this_prof_xnode) < 1
                qs = np.percentile(this_prof_res[this_prof_xnode_ind], q=[25, 50, 75])
                this_prof_med = qs[1]
                this_prof_dev = (qs[2] - qs[0]) / 2
                this_prof_out |= this_prof_xnode_ind & (np.abs(this_prof_res - 0) < ndev * this_prof_dev)
            else:
                this_prof_xnode_ind = np.abs(ap_im_xx_cor[istart:istart + chunk_len] - this_prof_xnode) < 1
                qs = np.percentile(this_prof_res[this_prof_xnode_ind], q=[25, 50, 75])
                this_prof_med = qs[1]
                this_prof_dev = (qs[2] - qs[0]) / 2
                this_prof_out |= this_prof_xnode_ind & (np.abs(this_prof_res - this_prof_med) < ndev * this_prof_dev)
        # second smooth --> as initial profile
        prof_init = SmoothSpline(
            ap_im_xx_cor[istart:istart + chunk_len][this_prof_out].flatten(),
            ap_im_norm[istart:istart + chunk_len][this_prof_out].flatten(),
            p=profile_smoothness)(prof_x)
        # append result
        profs.append(prof_init)
        prof_out[istart:istart + chunk_len] = this_prof_out

    # head and tail
    profs.insert(0, profs[0])
    profs.append(profs[-1])
    profs = np.array(profs)
    profs_y = np.hstack((-1, chunk_centers, n_rows + 1))
    # force positive
    profs = np.where(profs <= 0, 0., profs)

    # 3. interpolate profiles along lambda
    # [assuming profile varies along dispersion axis]
    ap_y = np.arange(n_rows)
    profs_interp = np.array([
        interp1d(profs_y, profs[:, icol], kind="linear")(ap_y) \
        for icol in range(profs.shape[1])]).T

    # 4. interpolate profiles along x
    profs_rebin = np.array([interp1d(prof_x, profs_interp[i], kind="linear", bounds_error=False, fill_value=0)(ap_im_xx_cor[i]) for i in range(n_rows)])
    # normalize
    profs_rebin = profs_rebin / np.sum(profs_rebin, axis=1).reshape(-1, 1)

    return profs_rebin, prof_x, profs_interp, prof_out


def extract_from_profile(ap_im, prof_recon, var=None):
    """ given profile, extract spectrum by lsq fitting

    Parameters
    ----------
    ap_im : ndarray
        The target image.
    prof_recon : ndarray
        The reconstructed profile.
    var : None or ndarray, optional
        The variance of ap_im, (1, npix_section) or like ap_im. The default is None.

    Returns
    -------
    spec_extr1 : ndarray
        The extracted spectrum from fitting profile.

    """
    if var is not None:
        # variance weighted extraction
        spec_extr1 = np.nansum(ap_im * prof_recon / var, axis=1) / np.nansum(prof_recon ** 2. / var, axis=1)
    else:
        # even weighted extraction (prefered)
        spec_extr1 = np.nansum(ap_im * prof_recon, axis=1) / np.nansum(prof_recon ** 2., axis=1)

    return spec_extr1


def extract_aperture(im, ap_center_interp, n_chunks=8, ap_width=15,
                     profile_oversample=10, profile_smoothness=1e-2,
                     num_sigma_clipping=5., gain=1., ron=0):
    """ extract an aperture given the aperture center

    Parameters
    ----------
    im : ndarray
        The target image.
    ap_center_interp : ndarray
        The ap_center_interp.
    n_chunks : int, optional
        The number of chunks. The default is 8.
    ap_width : float, optional
        The width of aperture / pix. The default is 15.
    profile_oversample : int, optional
        The oversampling factor of the profile. The default is 10.
    profile_smoothness : float, optional
        The smoothness of the profile. The default is 1e-2.
    num_sigma_clipping : float, optional
        The gaussian_sigma clipping threshold. The default is 5..
    gain : float, optional
        The gain of CCD. The default is 1..
    ron : flaot, optional
        The readout noise of CCD. The default is 0.

    Returns
    -------
    a dict consisting of many results.

    """

    # 1. get aperture section
    ap_im, ap_im_xx, ap_im_yy, ap_im_xx_cor = get_aperture_section(
        im, ap_center_interp, ap_width=ap_width)

    # set negative values to 0
    ap_im = np.where(ap_im > 0, ap_im, 0)

    # error image
    ap_im_err = np.sqrt(ap_im / gain + ron ** 2.)

    # 2. extract profile (quite good so far) for each chunk
    prof_recon, prof_xoversample, prof_oversample, prof_out = extract_profile(
        ap_im, ap_im_xx_cor, profile_smoothness=profile_smoothness,
        n_chunks=n_chunks, ap_width=ap_width,
        profile_oversample=profile_oversample)

    # 3. extract using profile
    spec_extr1 = extract_from_profile(ap_im, prof_recon, var=None)
    err_extr1 = extract_from_profile(ap_im_err, prof_recon, var=None)

    # 4. 3-gaussian_sigma clipping
    # reconstruct image
    ap_im_recon1 = prof_recon * spec_extr1.reshape(-1, 1)
    # residual
    ap_im_res = ap_im - ap_im_recon1
    # mask
    ap_im_mask = ap_im_res < ap_im_err * num_sigma_clipping
    prof_recon_masked = prof_recon * ap_im_mask

    # 5. re-extract using profile (robust but not always good)
    spec_extr2 = extract_from_profile(ap_im, prof_recon_masked, var=None)
    err_extr2 = extract_from_profile(ap_im_err, prof_recon_masked, var=None)

    # 6. combine extraction (robust)
    mask_extr = np.abs((spec_extr2 - spec_extr1) / err_extr2) > 3.
    spec_extr = np.where(mask_extr, spec_extr2, spec_extr1)
    err_extr = np.where(mask_extr, err_extr2, err_extr1)

    # reconstruct image
    ap_im_recon2 = prof_recon * spec_extr2.reshape(-1, 1)
    ap_im_recon = prof_recon * spec_extr.reshape(-1, 1)

    return dict(
        # ----- combined extraction -----
        spec_extr=spec_extr,
        err_extr=err_extr,
        # ----- first extraction -----
        spec_extr1=spec_extr1,
        err_extr1=err_extr1,
        # ----- second extraction (after gaussian_sigma-clipping) -----
        spec_extr2=spec_extr2,
        err_extr2=err_extr2,
        # ----- inconsistent pixels between 1&2, True if difference > 3sigma -----
        mask_extr=mask_extr,
        # ----- simple extraction -----
        spec_sum=ap_im.sum(axis=1),
        err_sum=ap_im_err.sum(axis=1),
        # ----- reconstructed profile -----
        prof_recon=prof_recon,
        # ----- reconstructed aperture -----
        ap_im=ap_im,
        ap_im_recon=ap_im_recon,
        ap_im_recon1=ap_im_recon1,
        ap_im_recon2=ap_im_recon2,
        # ----- aperture coordinates -----
        ap_im_xx=ap_im_xx,
        ap_im_xx_cor=ap_im_xx_cor,
        ap_im_yy=ap_im_yy,
        ap_im_out=prof_out,
        # ----- oversampled profile ------
        prof_xoversample=prof_xoversample,
        prof_oversample=prof_oversample,
    )


def extract_sum(ap_im, ap_im_xx_cor, ap_width=15):
    """ extract spectrum using simple sum

    Parameters
    ----------
    ap_im : ndarray
        The aperture image.
    ap_im_xx_cor : ndarray
        The x offsets.
    ap_width : float, optional
        The width of aperture. The default is 15.

    Returns
    -------
    ndarray
        simple extraction.

    """
    # determine fraction of pixels
    ap_im_frac = np.ones_like(ap_im)
    # fractional part
    ap_im_frac_l = ap_im_xx_cor + (.5 + ap_width)
    ap_im_frac_r = (.5 + ap_width) - ap_im_xx_cor
    ap_im_frac = np.where(ap_im_frac_l < 1, ap_im_frac_l, ap_im_frac)
    ap_im_frac = np.where(ap_im_frac_r < 1, ap_im_frac_r, ap_im_frac)
    # eliminate negative values
    ap_im_frac = np.where(ap_im_frac < 0, 0, ap_im_frac)

    # sum
    return np.nansum(ap_im_frac * ap_im, axis=1)


####################################
# extract all apertures
####################################

def extract_all(im, ap, n_chunks=8, profile_oversample=10, profile_smoothness=1e-2,
                num_sigma_clipping=10, gain=1., ron=0, n_jobs=-1,
                verbose=False, backend="multiprocessing"):
    """ extract all apertures with both simple & profile extraction

    Parameters
    ----------
    im : ndarray
        The target image.
    ap : Aperture
        The Aperture instance.
    n_chunks : int, optional
        The number of chunks. The default is 8.
    profile_oversample : int, optional
        The oversampling factor of the profile. The default is 10.
    profile_smoothness : float, optional
        The smoothness of the profile. The default is 1e-2.
    num_sigma_clipping : float, optional
        The gaussian_sigma clipping threshold. The default is 5..
    gain : float, optional
        The gain of CCD. The default is 1..
    ron : flaot, optional
        The readout noise of CCD. The default is 0.
    n_jobs : int, optional
        The number of processes launched. The default is -1.
    verbose :
        joblib verbose
    backend :
        joblib backend

    Returns
    -------
    dict
        a dict sconsisting of many results.

    """
    # extract all apertures in parallel
    rs = joblib.Parallel(n_jobs=n_jobs, verbose=verbose, backend=backend)(joblib.delayed(extract_aperture)(
        im, ap.ap_center_interp[i], n_chunks=n_chunks, profile_oversample=profile_oversample,
        profile_smoothness=profile_smoothness, num_sigma_clipping=num_sigma_clipping, gain=gain, ron=ron)
                                                   for i in range(ap.nap))
    # reconstruct results
    result = dict(
        # combbined
        spec_extr=np.array([rs[i]["spec_extr"] for i in range(ap.nap)]),
        err_extr=np.array([rs[i]["err_extr"] for i in range(ap.nap)]),
        # profile extraction
        spec_extr1=np.array([rs[i]["spec_extr1"] for i in range(ap.nap)]),
        err_extr1=np.array([rs[i]["err_extr1"] for i in range(ap.nap)]),
        # profile extraction gaussian_sigma-clipping
        spec_extr2=np.array([rs[i]["spec_extr2"] for i in range(ap.nap)]),
        err_extr2=np.array([rs[i]["err_extr2"] for i in range(ap.nap)]),
        # simple extraction
        spec_sum=np.array([rs[i]["spec_sum"] for i in range(ap.nap)]),
        err_sum=np.array([rs[i]["err_sum"] for i in range(ap.nap)]),
        # 1 for difference > 3 gaussian_sigma
        mask_extr=np.array([rs[i]["mask_extr"] for i in range(ap.nap)]),
    )
    return result


####################################
# make normalized FLAT image
####################################

def local_filter1(x, kw=5, method="mean"):
    """ 1d mean/median filter, used to smooth blaze function """
    if method == "mean":
        f = np.mean
    elif method == "median":
        f = np.median
    else:
        raise(ValueError("bad value for method!"))

    xs = np.copy(x)
    for i in range(kw, np.int(len(x) - kw)):
        xs[i] = f(x[np.int(i - kw):np.int(i + kw + 1)])
    return xs


def make_normflat(im, ap, max_dqe=0.04, min_snr=20, smooth_blaze=5, n_chunks=8,
                  profile_oversample=10, profile_smoothness=1e-2, num_sigma_clipping=20, gain=1., ron=0,
                  n_jobs=-1, verbose=False, backend="multiprocessing"):
    """ normalize FLAT

    Parameters
    ----------
    im : ndarray
        The target image.
    ap : Aperture
        The aperture instance
    max_dqe : float, optional
        The max deviation of Quantum Efficiency from 1.0. The default is 0.04.
    min_snr : flaot, optional
        Ignore the region with snr<min_snr. The default is 20.
    smooth_blaze : int, optional
        The smooth kernel width / pixel. The default is 5.
    n_chunks : int, optional
        Split each aperture to n_chunks chunks. The default is 8.
    profile_oversample : int, optional
        Oversampling factor of spatial profile. The default is 10.
    profile_smoothness : float, optional
        The smoothness of the profile. The default is 1e-2.
    num_sigma_clipping : float, optional
        The gaussian_sigma-clipping value / gaussian_sigma. The default is 20.
    gain : flaot, optional
        The gain of the image. The default is 1..
    ron : float, optional
        The readout noise. The default is 0.
    n_jobs : int, optional
        The number of processes launched. The default is -1.
    verbose:
        defaults to False
    backend:
        defaults to multiprocessing

    Returns
    -------
    blaze, im_norm : ndarray
        The blaze functions and sensitivity image.


    """
    im_recon = np.zeros_like(im)
    blaze = []

    # extract in parallel
    rs = joblib.Parallel(n_jobs=n_jobs, verbose=verbose, backend=backend)(joblib.delayed(extract_aperture)(
        im, ap.ap_center_interp[i], n_chunks=n_chunks, ap_width=ap.ap_width,
        profile_oversample=profile_oversample,
        profile_smoothness=profile_smoothness,
        num_sigma_clipping=num_sigma_clipping,
        gain=gain, ron=ron) for i in range(ap.nap))
    # gather results
    for i in range(ap.nap):
        # smooth blaze function (use spec_extr1???)
        this_blaze_smoothed1 = local_filter1(rs[i]["spec_extr1"], kw=np.int(smooth_blaze), method="median")
        this_blaze_smoothed2 = local_filter1(this_blaze_smoothed1, kw=np.int(smooth_blaze), method="mean")
        blaze.append(this_blaze_smoothed2)
        im_recon[rs[i]["ap_im_yy"], rs[i]["ap_im_xx"]] = rs[i]["prof_recon"] * this_blaze_smoothed2.reshape(-1, 1)
    blaze = np.array(blaze)
    # make norm image
    im_norm = im / im_recon

    # max deviation for quantum efficiency
    im_norm = np.where(np.abs(im_norm - 1) < max_dqe, im_norm, 1.)

    # for low SNR pixel, set 1
    im = np.where(im <= 0, 0, im)
    im_err = np.sqrt(im / gain + ron ** 2)
    im_snr = im / im_err
    im_norm = np.where(im_snr < min_snr, 1, im_norm)

    return blaze, im_norm


"""
Wavelength calibration, possibly flux calibration in the future.
"""


class Calibrator:
    def __init__(self):
        pass

    def cal_wave(self):
        pass

    def cal_flux(self):
        raise NotImplementedError("This is not implemented in present.")




import matplotlib.pyplot as plt
import numpy as np

xx = np.array(
    [
        [1, 2],
        [2, 3]
    ]
)

yy = np.array(
    [
        [1, 1],
        [2, 2]
    ]
)

zz = np.array(
    [
        [1, 2],
        [3, 4]
    ]
)

fig = plt.figure()
ax = fig.add_subplot(projection='3d')
ax.scatter(xx, yy, zz, s=10, c="cyan", label="Data")
ax.plot_surface(
    xx, yy, zz, rstride=1, cstride=1, cmap=plt.cm.jet,
    linewidth=0, antialiased=False, alpha=.5, label="Poly2dFitter"
)
# ax.legend()
ax.set_xlabel("X")
ax.set_ylabel("Y")
ax.set_zlabel("Z")
fig.tight_layout()



import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from tqdm import tqdm
from songcn.core.polynomial import PolyFit1D
from songcn.core.trace import trace_one_aperture, find_local_max


class ApertureList(list):
    def __init__(self, *args, **kwargs):
        super(ApertureList, self).__init__(*args, **kwargs)

    def __repr__(self):
        return f"<ApertureList: N={len(self)}>"

    @property
    def ap_center(self):
        return np.array([_.ap_center for _ in self])

    def get_mask(self, ap_width=None):
        if len(self) == 0:
            return False
        else:
            mask = self[0].get_mask(ap_width=ap_width)
            for _ in self[1:]:
                mask |= _.get_mask(ap_width=ap_width)
            return mask

    def view_profile(self, img, row=0, fov_width=20, norm=True):
        plt.figure()
        for ap in self:
            ap_center = ap.ap_center[row]
            ap_center_floor = ap.ap_center_floor[row]
            ap_coord = np.arange(-fov_width, fov_width + 1)
            plt.plot(
                ap_coord,
                img[row, ap_center_floor - fov_width:ap_center_floor + fov_width + 1] / img[row, ap_center_floor]
            )

    def clip(self):
        """ Reserve the apertures is_good and in_bounds. """
        for idx in range(len(self))[::-1]:
            if not (self[idx].is_good and self[idx].in_bounds):
                self.pop(idx)


class Aperture:
    def __init__(self, imgshape, ap_col, ap_mask=None, ap_width=15, deg=4):
        # the image shape is used
        self.n_row, self.n_col = imgshape
        assert self.n_row == len(ap_col)
        self.ap_coord = np.arange(len(ap_col), dtype=int)
        self.ap_col = ap_col
        self.ap_width = ap_width
        self.ap_mask = np.zeros_like(ap_col, dtype=bool) if ap_mask is None else ap_mask
        self.is_good = sum(self.ap_mask) == 0
        self.pf1 = PolyFit1D(
            self.ap_coord[~self.ap_mask],
            self.ap_col[~self.ap_mask],
            deg=deg, pw=1, robust=True
        )
        self.ap_center = self.pf1(self.ap_coord)
        self.ap_center_floor = np.floor(self.ap_center).astype(int)
        self.ap_center_remainder = self.ap_center - self.ap_center_floor
        self.ap_lower = self.ap_center - self.ap_width
        self.ap_upper = self.ap_center + self.ap_width

        self.in_bounds = (
                all(self.ap_lower > 0) and
                all(self.ap_lower < self.n_row - 1) and
                all(self.ap_upper > 0) and
                all(self.ap_upper < self.n_row - 1)
        )

    @property
    def ap_region_col(self):
        return self.ap_center_floor.reshape(-1, 1) - self.ap_width + np.arange(2 * self.ap_width + 1)

    @property
    def ap_region_row(self):
        return np.repeat(self.ap_coord.reshape(-1, 1), 2 * self.ap_width + 1, axis=1)

    def get_cutout(self, img, ind_row=(0, 256)):
        slc = slice(*ind_row)
        return self.ap_region_row[slc], \
               self.ap_region_col[slc], \
               img[self.ap_region_row, self.ap_region_col][slc]

    def get_cutout_corrected(self, img, ind_row=(0, 256)):
        slc = slice(*ind_row)
        return self.ap_region_row[slc], \
            self.ap_region_col[slc] - self.ap_center[slc].reshape(-1, 1), \
            img[self.ap_region_row, self.ap_region_col][slc]

    def get_mask(self, ap_width=None):
        mask = np.zeros((self.n_row, self.n_col), dtype=bool)
        if ap_width is None:
            mask[self.ap_region_row, self.ap_region_col] = True
        else:
            ap_region_row = np.repeat(self.ap_coord.reshape(-1, 1), 2 * ap_width + 1, axis=1)
            ap_region_col = self.ap_center_floor.reshape(-1, 1) - ap_width + np.arange(2 * ap_width + 1)
            mask[ap_region_row, ap_region_col] = True
        return mask

    def __repr__(self):
        return f"<Aperture: ({self.n_row}, {self.n_col}): is good: {self.is_good}: in bounds:{self.in_bounds}>"

    @property
    def region(self):
        return


def test():
    fp = "/Users/cham/projects/song/star_spec/20191105/night/ext/masterflat_20191105_slit5.fits"
    img = fits.getdata(fp)
    central_ind = int(img.shape[0] / 2)
    central_slice = img[central_ind:central_ind + 10].sum(axis=0)
    localmax = find_local_max(arr=central_slice, kernel_width=15)

    import logging
    logging.info("Start tracing apertures ...")
    al = ApertureList()
    for i_ap in tqdm(range(localmax.num), unit="Aperture"):
        # print(f"Tracing Aperture {i_ap}")
        ap_col, ap_mask = trace_one_aperture(
            img, init_pos=(central_ind, localmax.max_ind[i_ap]), extra_bin=1, kernel_width=15, maxdev=4)
        ap = Aperture(img.shape, ap_col, ap_mask, ap_width=15, deg=2)
        al.append(ap)
    al.clip()
    print(al, al.get_mask().sum())
    fig, axs = plt.subplots(1, 2, figsize=(8, 4))
    axs[0].imshow(np.log10(img), origin="lower")
    axs[1].imshow(al.get_mask(10), origin="lower")

    # al.view_profile(img, 1)  # 11 is good for estimating background

    # remove background
    from songcn.core.background import background

    bg = background(img, al, q=(30, 3), median_sigma=11, gaussian_sigma=11)
    img_bg = img - bg

    # extract spectrum
    al.view_profile(img_bg, 1000)  # 11 is good for estimating background

    fig, axs = plt.subplots(1, 2, figsize=(12, 6), sharex=True, sharey=True)
    axs[0].imshow(np.log10(img))
    axs[1].imshow(np.log10(bg))
    axs[0].plot(al.ap_center.T, np.arange(2048), "w-")
    axs[1].plot(al.ap_center.T, np.arange(2048), "w-")


# ap = al[30]
#
# ap_row, ap_col, ap_img = ap.get_cutout(img_bg)
#
# figure()
# imshow(ap_row, ap_col+ap.ap_center_remainder[:256][:, None],  ap_img)
#
# #%%
# ap_row, ap_col, ap_img = ap.get_cutout(img_bg, (0, 100))
# from matplotlib.colors import LightSource
# import matplotlib.pyplot as plt
# # Set up plot
# fig, ax = plt.subplots(subplot_kw=dict(projection='3d'))
#
# ls = LightSource(270, 45)
# # To use a custom hillshading mode, override the built-in shading and pass
# # in the rgb colors of the shaded surface calculated from "shade".
# rgb = ls.shade(ap_img, cmap=cm.jet, vert_exag=0.1, blend_mode='soft')
# surf = ax.plot_surface(ap_row, ap_col, ap_img, rstride=1, cstride=1, facecolors=rgb,
#                        linewidth=0, antialiased=False, shade=False)
#
